from collections.abc import Generator
from typing import NoReturn
from typing import Optional

from .tree import AST


class Pipeline:

    def __init__(self) -> NoReturn:
        self._ast = AST()

    def parse(self, pipeline: str) -> list:
        return self._ast.parse(pipeline)

    def search(self, key: str, pipeline: str) -> Generator[tuple, None, None]:
        key = key.split(".")
        # TODO ?
        if key[-1] == "*":
            msg = "Wildcard is not supported at the end of the key."
            raise ValueError(msg)

        for element in self._ast.parse(pipeline):
            yield from Pipeline._recursive_search(key, element)

    # thx Francois Garillot
    # https://stackoverflow.com/a/8848959
    @staticmethod
    def _matcher(_l1: list, _l2: list) -> bool:
        if not _l1:
            return _l2 == [] or _l2 == ["*"]
        if _l2 == [] or _l2[0] == "*":
            return Pipeline._matcher(_l2, _l1)
        if _l1[0] == "*":
            return Pipeline._matcher(_l1, _l2[1:]) or Pipeline._matcher(_l1[1:], _l2)
        if _l1[0] == _l2[0]:
            return Pipeline._matcher(_l1[1:], _l2[1:])
        else:
            return False

    @staticmethod
    def _recursive_search(key: list, _element: list, _actual_key: Optional[list] = None) -> list[tuple[str, str | int | float | list]]:

        if _actual_key is None:
            _actual_key = []

        _actual_key = _actual_key.copy()
        _matched = []

        if isinstance(_element, list):

            if len(_element) == 2 and isinstance(_element[0], str):
                _actual_key.append(_element[0])

            if Pipeline._matcher(_actual_key, key):
                _matched.append((".".join(_actual_key), _element[1]))
            else:
                for _sub in _element:
                    if not isinstance(_sub, list):
                        continue
                    _matched += Pipeline._recursive_search(key, _sub, _actual_key)

        return _matched



import unittest
from typing import NoReturn

from logstash_pipeline_parser import Pipeline


class PipelineTestCase(unittest.TestCase):

    def setUp(self) -> NoReturn:
        self.ppl = Pipeline()

    def test_parse(self) -> NoReturn:
        self.assertEqual(self.ppl.parse("""
            input {
              syslog {
                port => 123
                codec => cef
                qhashmap => {
                  somekey => "vaur"
                }
              }
            }
              """), [["input", [["syslog", [["port", 123], ["codec", "cef"], ["qhashmap", [["somekey", "vaur"]]]]]]]])

    def test_matcher(self) -> NoReturn:
        self.assertTrue(Pipeline._matcher(["input", "syslog", "port"], ["input", "syslog", "port"]))
        self.assertTrue(Pipeline._matcher(["input", "syslog", "port"], ["*", "syslog", "port"]))
        self.assertTrue(Pipeline._matcher(["input", "syslog", "port"], ["input", "*", "port"]))
        self.assertTrue(Pipeline._matcher(["input", "syslog", "port"], ["input", "syslog", "*"]))
        self.assertTrue(Pipeline._matcher(["input", "syslog", "udp", "port"], ["input", "*", "port"]))
        self.assertTrue(Pipeline._matcher(["input", "syslog", "port"], ["*"]))

    def test_search(self) -> NoReturn:
        data = """
            input {
              syslog {
                port => 123
                codec => cef
                some_key => {
                  somekey => "value"
                }
              }

              udp {
                sub => {
                  port => 456
                  host => "0.0.0.0"
                }
              }
            }

            filter {
              plug {
                port => 789 # comment
                somearr => ["abc", "def"]
              }
                # comment
              if [se][lec][tor] < 2 {
                plugin1 {
                  name_1 => 42
                }
              }
              else if 1 < 2 {
                plugin2 {
                  name_1 => 42
                }
              }
              # some comment
              else {
                plugin3 {
                  name_1 => 42
                }
              }
            }
            """

        with self.assertRaises(ValueError):
            list(self.ppl.search("some.key.*", data))

        self.assertEqual(
            list(self.ppl.search("input.syslog.port", data)),
            [("input.syslog.port", 123)]
        )

        self.assertEqual(
            list(self.ppl.search("input.*.port", data)),
            [("input.syslog.port", 123), ("input.udp.sub.port", 456)]
        )

        self.assertEqual(
            list(self.ppl.search("*.port", data)),
            [("input.syslog.port", 123), ("input.udp.sub.port", 456), ("filter.plug.port", 789)]
        )

        self.assertEqual(
            list(self.ppl.search("*.name_1", data)),
            [("filter.if.plugin1.name_1", 42), ("filter.else if.plugin2.name_1", 42), ("filter.else.plugin3.name_1", 42)]
        )

        self.assertEqual(
            list(self.ppl.search("filter.plug.somearr", data)),
            [("filter.plug.somearr", ["abc", "def"])]
        )

        self.assertEqual(
            list(self.ppl.search("input.udp", data)),
            [("input.udp", [["sub", [["port", 456], ["host", "0.0.0.0"]]]])]
        )


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import NoReturn

from pyparsing import pyparsing_test as ppt

from logstash_pipeline_parser.tree import AST


class ASTTestCase(ppt.TestParseResultsAsserts, unittest.TestCase):

    def test_comment(self) -> NoReturn:
        self.assertParseAndCheckList(AST.comment, "# comment", ["# comment"])

    def test_double_quoted_string(self) -> NoReturn:
        # grok { match => { "message" => " ... \"%{GREEDYDATA:data}\"" }
        self.assertParseAndCheckList(AST.double_quoted_string, r'"st\"ring\""', ['st"ring"'])
        self.assertParseAndCheckList(AST.double_quoted_string, r'"string"', ["string"])

    def test_single_quoted_string(self) -> NoReturn:
        self.assertParseAndCheckList(AST.single_quoted_string, r"'st\'ring\''", ["st'ring'"])
        self.assertParseAndCheckList(AST.single_quoted_string, "'string'", ["string"])

    def test_compare_operator(self) -> NoReturn:
        self.assertParseAndCheckList(AST.compare_operator, "==", ["=="])
        self.assertParseAndCheckList(AST.compare_operator, "!=", ["!="])
        self.assertParseAndCheckList(AST.compare_operator, "<=", ["<="])
        self.assertParseAndCheckList(AST.compare_operator, ">=", [">="])
        self.assertParseAndCheckList(AST.compare_operator, "<", ["<"])
        self.assertParseAndCheckList(AST.compare_operator, ">", [">"])

    def test_boolean_operator(self) -> NoReturn:
        self.assertParseAndCheckList(AST.boolean_operator, "and", ["and"])
        self.assertParseAndCheckList(AST.boolean_operator, "or", ["or"])
        self.assertParseAndCheckList(AST.boolean_operator, "xor", ["xor"])
        self.assertParseAndCheckList(AST.boolean_operator, "nand", ["nand"])

    def test_regexp_operator(self) -> NoReturn:
        self.assertParseAndCheckList(AST.regexp_operator, "=~", ["=~"])
        self.assertParseAndCheckList(AST.regexp_operator, "!~", ["!~"])

    def test_in_operator(self) -> NoReturn:
        self.assertParseAndCheckList(AST.in_operator, "in", ["in"])

    def test_not_in_operator(self) -> NoReturn:
        self.assertParseAndCheckList(AST.not_in_operator, "not in", ["not", "in"])
        self.assertParseAndCheckList(AST.not_in_operator, "not     in", ["not", "in"])
        self.assertParseAndCheckList(AST.not_in_operator, "not\tin", ["not", "in"])

    def test_string(self) -> NoReturn:
        self.assertParseAndCheckList(AST.string, "'string'", ["string"])
        self.assertParseAndCheckList(AST.string, '"string"', ["string"])

    def test_bare_word(self) -> NoReturn:
        self.assertParseAndCheckList(AST.bare_word, "a0bc_", ["a0bc_"])
        self.assertParseAndCheckList(AST.bare_word, "_a0bc", ["_a0bc"])

        with self.assertRaisesParseException():
            self.assertParseAndCheckList(AST.bare_word, "0abc", [])

    def test_method(self) -> NoReturn:
        self.assertParseAndCheckList(AST.method, "a0bc_", ["a0bc_"])
        self.assertParseAndCheckList(AST.method, "_a0bc", ["_a0bc"])

        with self.assertRaisesParseException():
            self.assertParseAndCheckList(AST.method, "0abc", [])

    def test_number(self) -> NoReturn:
        self.assertParseAndCheckList(AST.number, "42", [42])
        self.assertParseAndCheckList(AST.number, "-42", [-42])
        self.assertParseAndCheckList(AST.number, "42.42", [42.42])
        self.assertParseAndCheckList(AST.number, "-42.42", [-42.42])

    def test_regexp(self) -> NoReturn:
        self.assertParseAndCheckList(AST.regexp, "/rgx/", ["rgx"])

    def test_plugin_type(self) -> NoReturn:
        self.assertParseAndCheckList(AST.plugin_type, "input", ["input"])
        self.assertParseAndCheckList(AST.plugin_type, "filter", ["filter"])
        self.assertParseAndCheckList(AST.plugin_type, "output", ["output"])

    def test_name(self) -> NoReturn:
        self.assertParseAndCheckList(AST.name, "word_1", ["word_1"])
        self.assertParseAndCheckList(AST.name, "'string'", ["string"])
        self.assertParseAndCheckList(AST.name, '"string"', ["string"])

    def test_selector_element(self) -> NoReturn:
        self.assertParseAndCheckList(AST.selector_element, "[string]", ["[", "string", "]"])

        with self.assertRaisesParseException():
            self.assertParseAndCheckList(AST.selector_element, "[str,ing]", [])

    def test_selector(self) -> NoReturn:
        self.assertParseAndCheckList(AST.selector, "[str1][str2][str3]", ["[str1][str2][str3]"])

    def test_hash_entry(self) -> NoReturn:
        self.assertParseAndCheckList(AST.hash_entry, "42 => 42", [[42, 42]])
        self.assertParseAndCheckList(AST.hash_entry, "42    => \t 42", [[42, 42]])
        self.assertParseAndCheckList(AST.hash_entry, "bareword => bareword", [["bareword", "bareword"]])
        self.assertParseAndCheckList(AST.hash_entry, "'string' => 'string'", [["string", "string"]])
        self.assertParseAndCheckList(AST.hash_entry, '"string" => "string"', [["string", "string"]])

    def test_hash_entries(self) -> NoReturn:
        self.assertParseAndCheckList(AST.hash_entries, "name1 => value1 name2 => value2", [[["name1", "value1"], ["name2", "value2"]]])

    def test_hashmap(self) -> NoReturn:
        self.assertParseAndCheckList(AST.hashmap, "{ name1 => value1 }", [[["name1", "value1"]]])
        self.assertParseAndCheckList(AST.hashmap, """
        {
        # comment
        name1 => value1 # comment
        # comment
        }
        """, [[["name1", "value1"]]])

    def test_array(self) -> NoReturn:
        self.assertParseAndCheckList(AST.array, "[ ]", [[]])
        self.assertParseAndCheckList(AST.array, "[ 42,     bare_word, \t 'string' ]", [[42, "bare_word", "string"]])
        self.assertParseAndCheckList(AST.array, "[ 42, bare_word, [ sub_array, 'value' ] ]", [[42, "bare_word", ["sub_array", "value"]]])

    def test_attribute(self) -> NoReturn:
        self.assertParseAndCheckList(AST.attribute, "name_1 => value1", [["name_1", "value1"]])

    def test_compare_expression(self) -> NoReturn:
        self.assertParseAndCheckList(AST.compare_expression, "42 == 42", [[42, "==", 42]])
        self.assertParseAndCheckList(AST.compare_expression, "42 != 'string'", [[42, "!=", "string"]])
        self.assertParseAndCheckList(AST.compare_expression, "[one]    < \t [two]", [["[one]", "<", "[two]"]])

    def test_in_expression(self) -> NoReturn:
        self.assertParseAndCheckList(AST.in_expression, "2    in \t '42'", [[2, "in", "42"]])

    def test_not_in_expression(self) -> NoReturn:
        self.assertParseAndCheckList(AST.not_in_expression, "3   not  \t  in \t '42'", [[3, "not", "in", "42"]])

    def test_regexp_expression(self) -> NoReturn:
        self.assertParseAndCheckList(AST.regexp_expression, "[some][name]   =~ \t 'string'", ["[some][name]", "=~", "string"])
        self.assertParseAndCheckList(AST.regexp_expression, "[some][name]   =~ \t /rgx/", ["[some][name]", "=~", "rgx"])

    def test_method_call(self) -> NoReturn:
        self.assertParseAndCheckList(AST.method_call, "method_name  ( 'rval1', \t 42 )", ["method_name", ["rval1", 42]])

    def test_plugin(self) -> NoReturn:
        self.assertParseAndCheckList(AST.plugin, "plugin_name  { name_1 => 42 name_2 => bare_word }", [["plugin_name", [["name_1", 42], ["name_2", "bare_word"]]]])

    def test_plugin_section(self) -> NoReturn:
        self.assertParseAndCheckList(AST.plugin_section, "input { beats { port => 42 } }", [["input", [["beats", [["port", 42]]]]]])

    def test_condition(self) -> NoReturn:
        self.assertParseAndCheckList(AST.condition, "'a' in 'abc' and 'b'     not in 'ok' or      1 < 2 ", [["a", "in", "abc"], "and", ["b", "not", "in", "ok"], "or", [1, "<", 2]])

    def test_if_rule(self) -> NoReturn:
        self.assertParseAndCheckList(AST.if_rule, "if [some][true][value] { plugin_name { name_1 => 42 } }", [["if", ["[some][true][value]", [["plugin_name", [["name_1", 42]]]]]]])

    def test_else_if_rule(self) -> NoReturn:
        self.assertParseAndCheckList(AST.else_if_rule, "else   \t  if [some][true][value] { plugin_name { name_1 => 42 } }",
                                     [["else if", ["[some][true][value]", [["plugin_name", [["name_1", 42]]]]]]])
        self.assertParseAndCheckList(AST.else_if_rule, "else  \n#comment\r\n  if [some][true][value] { plugin_name { name_1 => 42 } }",
                                     [["else if", ["[some][true][value]", [["plugin_name", [["name_1", 42]]]]]]])

    def test_else_rule(self) -> NoReturn:
        self.assertParseAndCheckList(AST.else_rule, "else  \t  { plugin_name { name_1 => 42 } }", [["else", [["plugin_name", [["name_1", 42]]]]]])

    def test_negative_expression(self) -> NoReturn:
        self.assertParseAndCheckList(AST.negative_expression, "! [sel][ector]", [["!", "[sel][ector]"]])
        self.assertParseAndCheckList(AST.negative_expression, "! ( 1 < 2) ", [["!", [1, "<", 2]]])

    def test_branch(self) -> NoReturn:
        self.assertParseAndCheckList(AST.branch, """
        if 1 < 2 {
          plugin1 {
            name_1 => 42
          }
        }""", [["if", [[1, "<", 2], [["plugin1", [["name_1", 42]]]]]]])

        self.assertParseAndCheckList(AST.branch, """
        if 1 < 2 {
          plugin1 {
            name_1 => 42
          }
        }
        else {
          plugin2 {
            name_1 => 42
          }
        }""", [["if", [[1, "<", 2], [["plugin1", [["name_1", 42]]]]]], ["else", [["plugin2", [["name_1", 42]]]]]])

        self.assertParseAndCheckList(AST.branch, """
        if 1 < 2 {
          plugin1 {
            name_1 => 42
          }
        }

        else if [some][true][value] {
          plugin2 {
            name_1 => 42
          }
        }

        else {
          plugin3 {
            name_1 => 42
          }
        }""", [
            ["if", [
                [1, "<", 2], [["plugin1", [["name_1", 42]]]]
            ]],
            ["else if", [
                "[some][true][value]", [["plugin2", [["name_1", 42]]]]
            ]],
            ["else", [
                ["plugin3", [["name_1", 42]]]
            ]]
        ])

    def test_config(self) -> NoReturn:
        self.assertParseAndCheckList(AST.config, """
        input {
          plugin1 {
            name_1 => 42
          }
        }

        filter {
          plugin2 {
            name_1 => 42
          }
        }

        output {
          plugin3 {
            name_1 => 42
          }
        }""", [["input", [["plugin1", [["name_1", 42]]]]], ["filter", [["plugin2", [["name_1", 42]]]]], ["output", [["plugin3", [["name_1", 42]]]]]])

    def test_complex(self) -> NoReturn:
        self.assertParseAndCheckList(AST.config, r"""
            input {
                some_plugin {
                    config_mode => "advanced"
                    threads => 8
                    decorate_events => true # comment behind value
                    storage_connection => "DefaultEndpointsProtocol..."
                    # array
                    event_hubs => [
                        42,
                        "double quoted",
                        "single quoted",
                        # hashmap
                        {
                            # hashentry => hashmap
                            "insights-operational-logs" => {
                                # hashentry
                                event_hub_connection => "Endpoint=sb://example1..."
                                initial_position => "beginning"
                                consumer_group => "iam_team"
                            }
                        # trailing comma
                        },
                    ]
                }
            }

            filter {
                if [log][file][path] in ["/path/one", "/path/two"] {
                    grok {
                      match => { "original_message" => "\[(?<timestamp>%{YEAR}...\"%{GREEDYDATA:message}\""}
                    }
                    date {
                      match => [ "timestamp", "yyyy-MM-dd HH:mm:ss Z" ]
                    }
                }

                else if 1 < 42
                {
                    grok {
                      match => { "original_message" => "double quoted grok pattern"}
                    }

                    if "_grokparsefailure" not in [tags] {
                        date {
                            match => [ "timestamp", "ISO8601" ]
                        }
                    }
                }
                else {
                    mutate {
                        convert => {
                            "fieldname" => "integer"
                            "booleanfield" => "boolean"
                        }
                        copy => { "source_field" => "dest_field" }
                        gsub => [
                            # replace all forward slashes with underscore
                            "fieldname", "/", "_",
                            # replace backslashes, question marks, hashes, and minuses
                            # with a dot "."
                            "fieldname2", "[\\?#-]", "."
                        ]
                    }
                    mutate { add_field => { "[field][keyword]" => "TEST-SOMEDATA" } }

                    if [field][keyword] =~ /^(TEST|test)-.*$/ {
                        mutate { add_tag => [ "TEST" ] }
                    }
                }
            }

            output {
                file {
                    path => "/tmp/output.txt"
                    codec => line { format => "custom format: %{message}"}
                }
            }

            output {

                elasticsearch
                {
                    hosts => ["https://elasticnode01.local:9200", "https://elasticnode02.local:9200"]

                    index => "elk-filebeat-%{[agent][version]}-%{+YYYY}"
                    ssl => true
                    ssl_certificate_verification => true
                    keystore => "/etc/logstash/certs/czsrv-jerlogstash01.p12"
                    keystore_password => "${elasticsearch.keystore.password}"
                    truststore => "/etc/logstash/certs/czsrv-jerlogstash01.p12"
                    truststore_password => "${elasticsearch.keystore.password}"

                    user => "logstash"
                    password => "${elasticsearch.logstash.password}"

                    manage_template => false
                }
            }
        """, [["input",
               [["some_plugin",
                 [["config_mode", "advanced"],
                  ["threads", 8],
                  ["decorate_events", "true"],
                  ["storage_connection", "DefaultEndpointsProtocol..."],
                  ["event_hubs",
                   [42,
                    "double quoted",
                    "single quoted",
                    [["insights-operational-logs",
                      [["event_hub_connection", "Endpoint=sb://example1..."],
                       ["initial_position", "beginning"],
                       ["consumer_group", "iam_team"]]]]]]]]]],
              ["filter",
               [["if",
                 [["[log][file][path]", "in", ["/path/one", "/path/two"]],
                  [["grok",
                    [["match",
                      [["original_message",
                        '[(?<timestamp>%{YEAR}..."%{GREEDYDATA:message}"']]]]],
                   ["date", [["match", ["timestamp", "yyyy-MM-dd HH:mm:ss Z"]]]]]]],
                ["else if",
                 [[1, "<", 42],
                  [["grok",
                    [["match", [["original_message", "double quoted grok pattern"]]]]],
                   ["if",
                    [["_grokparsefailure", "not", "in", "[tags]"],
                     [["date", [["match", ["timestamp", "ISO8601"]]]]]]]]]],
                ["else",
                 [["mutate",
                   [["convert", [["fieldname", "integer"], ["booleanfield", "boolean"]]],
                    ["copy", [["source_field", "dest_field"]]],
                    ["gsub", ["fieldname", "/", "_", "fieldname2", "[\\?#-]", "."]]]],
                  ["mutate", [["add_field", [["[field][keyword]", "TEST-SOMEDATA"]]]]],
                  ["if",
                   ["[field][keyword]",
                    "=~",
                    "^(TEST|test)-.*$",
                    [["mutate", [["add_tag", ["TEST"]]]]]]]]]]],
              ["output",
               [["file",
                 [["path", "/tmp/output.txt"],
                  ["codec", ["line", [["format", "custom format: %{message}"]]]]]]]],
              ["output",
               [["elasticsearch",
                 [["hosts",
                   ["https://elasticnode01.local:9200", "https://elasticnode02.local:9200"]],
                  ["index", "elk-filebeat-%{[agent][version]}-%{+YYYY}"],
                  ["ssl", "true"],
                  ["ssl_certificate_verification", "true"],
                  ["keystore", "/etc/logstash/certs/czsrv-jerlogstash01.p12"],
                  ["keystore_password", "${elasticsearch.keystore.password}"],
                  ["truststore", "/etc/logstash/certs/czsrv-jerlogstash01.p12"],
                  ["truststore_password", "${elasticsearch.keystore.password}"],
                  ["user", "logstash"],
                  ["password", "${elasticsearch.logstash.password}"],
                  ["manage_template", "false"]]]]]])


if __name__ == '__main__':
    unittest.main()



from re import compile

import pyparsing as pp
from pyparsing import common as ppc


# DEBUG
# ------------------------------------------------------------------------
# pp.enable_diag(pp.Diagnostics.warn_multiple_tokens_in_named_alternation)
# pp.enable_diag(pp.Diagnostics.warn_ungrouped_named_tokens_in_collection)
# pp.enable_diag(pp.Diagnostics.warn_name_set_on_empty_Forward)
# pp.enable_diag(pp.Diagnostics.warn_on_multiple_string_args_to_oneof)
# pp.enable_diag(pp.Diagnostics.warn_on_match_first_with_lshift_operator)
# pp.enable_diag(pp.Diagnostics.warn_on_parse_using_empty_Forward)
# pp.enable_diag(pp.Diagnostics.warn_on_assignment_to_Forward)
#
# pp.enable_diag(pp.Diagnostics.enable_debug_on_named_expressions)
# pp.enable_all_warnings()
# pp.autoname_elements()
# ------------------------------------------------------------------------


class AST:
    array_start = pp.Suppress(pp.Literal("["))
    array_stop = pp.Suppress(pp.Literal("]"))
    object_start = pp.Suppress(pp.Literal("{"))
    object_stop = pp.Suppress(pp.Literal("}"))
    parenthesis_start = pp.Suppress(pp.Literal("("))
    parenthesis_stop = pp.Suppress(pp.Literal(")"))
    setter = pp.Suppress(pp.Literal("=>"))

    value = pp.Forward().set_name("value")
    rvalue = pp.Forward().set_name("rvalue")
    expression = pp.Forward().set_name("expression")
    branch_or_plugin = pp.Forward().set_name("branch_or_plugin")

    r"""
      rule whitespace
        [ \t\r\n]+ <LogStash::Config::AST::Whitespace>
      end
    """

    whitespace = pp.White()
    whitespace.set_name("whitespace")

    r"""
      rule comment
        (whitespace? "#" [^\r\n]* "\r"? "\n")+ <LogStash::Config::AST::Comment>
      end
    """
    comment = pp.Regex(compile(r"#.*"))
    comment.set_name("comment")

    r"""
      rule cs
        (comment / whitespace)* <LogStash::Config::AST::Whitespace>
      end
    """

    cs = comment | whitespace
    cs.set_name("cs")

    r"""
      rule single_quoted_string
        ( "'" ( "\\'" / !"'" . )* "'" <LogStash::Config::AST::String>)
      end
    """
    single_quoted_string = pp.QuotedString(quote_char="'", esc_char="\\", escQuote="\\\\")
    single_quoted_string.set_name("single_quoted_string")

    r"""
      rule double_quoted_string
        ( '"' ( '\"' / !'"' . )* '"' <LogStash::Config::AST::String>)
      end
    """

    double_quoted_string = pp.QuotedString(quote_char='"', esc_char="\\", escQuote="\\\\")
    double_quoted_string.set_name("double_quoted_string")

    r"""
      rule compare_operator 
        ("==" / "!=" / "<=" / ">=" / "<" / ">") 
        <LogStash::Config::AST::ComparisonOperator>
      end
    """

    compare_operator = pp.Literal("==") | pp.Literal("!=") | pp.Literal("<=") | pp.Literal(">=") | pp.Literal("<") | pp.Literal(">")
    compare_operator.set_name("compare_operator")

    r"""
      rule compare_expression
        rvalue cs compare_operator cs rvalue
        <LogStash::Config::AST::ComparisonExpression>
      end
    """

    compare_expression = pp.Group(rvalue + compare_operator + rvalue)
    compare_expression.set_name("compare_expression")

    r"""
      rule boolean_operator
        ("and" / "or" / "xor" / "nand")
        <LogStash::Config::AST::BooleanOperator>
      end
    """

    boolean_operator = pp.Literal("and") | pp.Literal("or") | pp.Literal("xor") | pp.Literal("nand")
    boolean_operator.set_name("boolean_operator")

    r"""
      rule regexp_operator
        ("=~" / "!~") <LogStash::Config::AST::RegExpOperator>
      end
    """

    regexp_operator = pp.Literal("=~") | pp.Literal("!~")
    regexp_operator.set_name("regexp_operator")

    r"""
      rule in_operator
        "in"
      end
    """

    in_operator = pp.Literal("in")

    r"""
      rule in_expression
        rvalue cs in_operator cs rvalue
        <LogStash::Config::AST::InExpression>
      end
    """

    in_expression = pp.Group(rvalue + in_operator + rvalue)
    in_expression.set_name("in_expression")

    r"""
      rule not_in_operator
        "not " cs "in"
      end
    """

    # TODO podminky jako group?

    not_in_operator = pp.Literal("not") + pp.Literal("in")

    r"""
      rule not_in_expression
        rvalue cs not_in_operator cs rvalue
        <LogStash::Config::AST::NotInExpression>
      end
    """

    not_in_expression = pp.Group(rvalue + not_in_operator + rvalue)
    not_in_expression.set_name("not_in_expression")

    r"""
      rule selector_element
        "[" [^\]\[,]+ "]"
        <LogStash::Config::AST::SelectorElement>
      end
    """
    # [key]
    selector_element = pp.Literal("[") + pp.Regex(compile(r"[^\]\[,]+")) + pp.Literal("]")
    selector_element.set_name("selector_element")

    r"""
      rule selector
        selector_element+
        <LogStash::Config::AST::Selector>
      end
    """
    # [some][logstash][key] => ['[', 'soe', ']', '[', 'logstash', ']', '[', 'key', ']'] => ['[some][logstash][key]']
    selector = pp.Combine(pp.OneOrMore(selector_element))
    selector.set_name("selector")

    r"""
      rule string
        double_quoted_string / single_quoted_string
      end
    """

    string = double_quoted_string | single_quoted_string
    string.set_name("string")

    r"""
      rule bareword
        [A-Za-z_] [A-Za-z0-9_]+
        <LogStash::Config::AST::Bareword>
      end
    """

    bare_word = pp.Word(pp.alphas + "_", pp.alphanums + "_")
    bare_word.set_name("bare_word")

    r"""
      rule method
        bareword
      end
    """
    method = pp.Word(pp.alphas + "_", pp.alphanums + "_")
    method.set_name("method")

    r"""
      rule number
        "-"? [0-9]+ ("." [0-9]*)?
        <LogStash::Config::AST::Number>
      end
    """
    number = ppc.number

    r"""
      rule regexp
        ( '/' ( '\/' / !'/' . )* '/'  <LogStash::Config::AST::RegExp>)
      end
    """

    regexp = pp.QuotedString(quote_char="/")
    regexp.set_name("regexp")

    r"""
      rule regexp_expression
        rvalue cs  regexp_operator cs (string / regexp)
        <LogStash::Config::AST::RegexpExpression>
      end
    """

    regexp_expression = rvalue + regexp_operator + pp.Or([string, regexp])
    regexp_expression.set_name("regexp_expression")
    r"""
      rule plugin_type
        ("input" / "filter" / "output")
      end
    """

    plugin_type = pp.Literal("input") | pp.Literal("filter") | pp.Literal("output")
    plugin_type.set_name("plugin_type")

    r"""
      rule name
        (
          ([A-Za-z0-9_-]+ <LogStash::Config::AST::Name>)
          / string
        )
      end
    """
    name = pp.Word(pp.alphanums + "_-") | string
    name.set_name("name")

    r"""
      rule attribute
        name cs "=>" cs value
        <LogStash::Config::AST::Attribute>
      end
    """

    attribute = pp.Group(name + setter + value)
    attribute.set_name("attribute")

    r"""
      rule plugin
        name cs "{"
          cs
          attributes:( attribute (whitespace cs attribute)*)?
          cs
        "}"
        <LogStash::Config::AST::Plugin>
      end
    """

    plugin = pp.Group(name + pp.Group(object_start + pp.ZeroOrMore(attribute) + object_stop))
    plugin.set_name("plugin")

    """
      rule array_value
        bareword / string / number / array / hash
      end
    """
    # DEFINED IN TREETOP BUT NOT USED

    """
      rule plugins
        (plugin (cs plugin)*)?
        <LogStash::Config::AST::Plugins>
      end
    """

    # DEFINED IN TREETOP BUT NOT USED

    r"""
      rule array
        "["
        cs
        (
          value (_ "," _ value)*
        )?
        _
        "]"
        <LogStash::Config::AST::Array>
      end
    """
    # [abc, def, ghi]
    array = pp.Group(array_start + pp.DelimitedList(pp.ZeroOrMore(value)).set_name("array_values") + array_stop)
    array.set_name("array")

    r"""
      rule hashentry
        name:(number / bareword / string) cs "=>" cs value
        <LogStash::Config::AST::HashEntry>
      end
    """

    hash_entry = pp.Group((number | bare_word | string) + setter + value)
    hash_entry.set_name("hash_entry")

    r"""
      rule hashentries
        hashentry (whitespace hashentry)*
        <LogStash::Config::AST::HashEntries>
      end
    """

    hash_entries = pp.Group(pp.ZeroOrMore(hash_entry))
    hash_entries.set_name("hash_entries")

    r"""
      rule hash
        "{"
          cs
          hashentries?
          cs
        "}"
        <LogStash::Config::AST::Hash>
      end

      https://www.elastic.co/guide/en/logstash/current/configuration-file-structure.html#hash
    """

    hashmap = object_start + hash_entries + object_stop
    hashmap.set_name("hashmap")

    r"""
      rule method_call
          method cs "(" cs
            (
              rvalue ( cs "," cs rvalue )*
            )?
          cs ")"
        <LogStash::Config::AST::MethodCall>
      end
    """

    method_call = method + pp.Group(parenthesis_start + pp.DelimitedList(pp.ZeroOrMore(rvalue)) + parenthesis_stop)
    method_call.set_name("method_call")

    r"""
      rule plugin_section
        plugin_type cs "{"
          cs (branch_or_plugin cs)*
        "}"
        <LogStash::Config::AST::PluginSection>
      end
    """

    plugin_section = pp.Group(plugin_type + object_start + pp.Group(pp.ZeroOrMore(branch_or_plugin)) + object_stop)
    plugin_section.set_name("plugin_selection")

    r"""
      rule condition
        expression (cs boolean_operator cs expression)*
        <LogStash::Config::AST::Condition>
      end
    """

    condition = expression + pp.ZeroOrMore(boolean_operator + expression)
    condition.set_name("condition")

    r"""
      rule if
        "if" cs condition cs "{" cs (branch_or_plugin cs)* "}"
        <LogStash::Config::AST::If>
      end
    """

    if_rule = pp.Group(pp.Keyword("if") + pp.Group(condition + object_start + pp.Group(pp.ZeroOrMore(branch_or_plugin)) + object_stop))
    if_rule.set_name("if")

    r"""
      rule else_if
        "else" cs "if" cs condition cs "{" cs ( branch_or_plugin cs)* "}"
        <LogStash::Config::AST::Elsif>
      end
    """
    # "else    \n #comment\n  if" as one string "else if"
    else_if_rule = pp.Group(pp.Combine(pp.Keyword("else") + pp.ZeroOrMore(cs).set_parse_action(pp.replace_with(" ")) + pp.Keyword("if")) + pp.Group(
        condition + object_start + pp.Group(pp.ZeroOrMore(branch_or_plugin)) + object_stop))
    else_if_rule.set_name("else_if")

    r"""
      rule else
        "else" cs "{" cs (branch_or_plugin cs)* "}"
        <LogStash::Config::AST::Else>
      end
    """

    else_rule = pp.Group(pp.Keyword("else") + pp.Group(object_start + pp.ZeroOrMore(branch_or_plugin) + object_stop))
    else_rule.set_name("else")

    r"""
      rule branch
        if (cs else_if)* (cs else)?
        <LogStash::Config::AST::Branch>
      end
    """

    branch = if_rule + pp.ZeroOrMore(else_if_rule) + pp.Optional(else_rule)
    branch.set_name("branch")

    r"""
      rule negative_expression
        (
            ("!" cs "(" cs condition cs ")")
          / ("!" cs selector)
        ) <LogStash::Config::AST::NegativeExpression>
      end
    """

    negative_expression = pp.Group(pp.Literal("!") + parenthesis_start + condition + parenthesis_stop) | \
                          pp.Group(pp.Literal("!") + selector)
    negative_expression.set_name("negative_expression")

    r"""
      rule expression
        (
            ("(" cs condition cs ")")
          / negative_expression
          / in_expression
          / not_in_expression
          / compare_expression
          / regexp_expression
          / rvalue
        ) <LogStash::Config::AST::Expression>
      end
    """

    expression << (
            pp.Group(parenthesis_start + condition + parenthesis_stop) |
            negative_expression |
            in_expression |
            not_in_expression |
            compare_expression |
            regexp_expression |
            rvalue
    )

    r"""
      rule value
        plugin / bareword / string / number / array / hash
      end
    """

    value << (plugin | bare_word | string | number | array | hashmap)

    r"""
      rule rvalue
        string / number / selector / array / method_call / regexp
      end
    """

    rvalue << (string | number | selector | array | method_call | regexp)

    r"""
      rule branch_or_plugin
        branch / plugin
      end
    """

    branch_or_plugin << (branch | plugin)

    """
      rule config
        cs plugin_section cs (cs plugin_section)* cs <LogStash::Config::AST::Config>
      end
    """

    config = pp.ZeroOrMore(plugin_section)
    config.ignore(comment)
    config.set_name("config")

    def parse(self, pipeline: str) -> list:
        return self.config.parse_string(pipeline).as_list()
